"""Tests editing the document using the API (replace etc.)

The tests in this module directly call the edit APIs on the TextArea rather
than going via bindings.

Note that more extensive testing for editing is done at the Document level.
"""

import pytest

from textual.app import App, ComposeResult
from textual.widgets import TextArea
from textual.widgets.text_area import EditResult, Selection

TEXT = """\
I must not fear.
Fear is the mind-killer.
Fear is the little-death that brings total obliteration.
I will face my fear.
"""

SIMPLE_TEXT = """\
ABCDE
FGHIJ
KLMNO
PQRST
UVWXY
Z
"""


class TextAreaApp(App):
    def compose(self) -> ComposeResult:
        text_area = TextArea()
        text_area.load_text(TEXT)
        yield text_area


async def test_insert_text_start_maintain_selection_offset():
    """Ensure that we can maintain the offset between the location
    an insert happens and the location of the selection."""
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.move_cursor((0, 5))
        text_area.insert("Hello", location=(0, 0))
        assert text_area.text == "Hello" + TEXT
        assert text_area.selection == Selection.cursor((0, 10))


async def test_insert_text_start():
    """The document is correctly updated on inserting at the start.
    If we don't maintain the selection offset, the cursor jumps
    to the end of the edit and the selection is empty."""
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.move_cursor((0, 5))
        text_area.insert("Hello", location=(0, 0), maintain_selection_offset=False)
        assert text_area.text == "Hello" + TEXT
        assert text_area.selection == Selection.cursor((0, 5))


async def test_insert_empty_string():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.load_text("0123456789")

        text_area.insert("", location=(0, 3))

        assert text_area.text == "0123456789"


async def test_replace_empty_string():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.load_text("0123456789")

        text_area.replace("", start=(0, 3), end=(0, 7))

        assert text_area.text == "012789"


@pytest.mark.parametrize(
    "cursor_location,insert_location,cursor_destination",
    [
        ((0, 3), (0, 2), (0, 4)),  # API insert just before cursor
        ((0, 3), (0, 3), (0, 4)),  # API insert at cursor location
        ((0, 3), (0, 4), (0, 3)),  # API insert just after cursor
        ((0, 3), (0, 5), (0, 3)),  # API insert just after cursor
    ],
)
async def test_insert_character_near_cursor_maintain_selection_offset(
    cursor_location,
    insert_location,
    cursor_destination,
):
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.load_text("012345")
        text_area.move_cursor(cursor_location)
        text_area.insert("X", location=insert_location)
        assert text_area.selection == Selection.cursor(cursor_destination)


@pytest.mark.parametrize(
    "cursor_location,insert_location,cursor_destination",
    [
        ((1, 0), (0, 0), (2, 0)),  # API insert before cursor row
        ((0, 0), (0, 0), (1, 0)),  # API insert right at cursor row
        ((0, 0), (1, 0), (0, 0)),  # API insert after cursor row
    ],
)
async def test_insert_newline_around_cursor_maintain_selection_offset(
    cursor_location,
    insert_location,
    cursor_destination
):
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.move_cursor(cursor_location)
        text_area.insert("X\n", location=insert_location)
        assert text_area.selection == Selection.cursor(cursor_destination)


async def test_insert_newlines_start():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.insert("\n\n\n")
        assert text_area.text == "\n\n\n" + TEXT
        assert text_area.selection == Selection.cursor((3, 0))


async def test_insert_newlines_end():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.insert("\n\n\n", location=(4, 0))
        assert text_area.text == TEXT + "\n\n\n"


async def test_insert_windows_newlines():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        # Although we're inserting windows newlines, the configured newline on
        # the Document inside the TextArea will be "\n", so when we check TextArea.text
        # we expect to see "\n".
        text_area.insert("\r\n\r\n\r\n")
        assert text_area.text == "\n\n\n" + TEXT


async def test_insert_old_mac_newlines():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.insert("\r\r\r")
        assert text_area.text == "\n\n\n" + TEXT


async def test_insert_text_non_cursor_location():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.insert("Hello", location=(4, 0))
        assert text_area.text == TEXT + "Hello"
        assert text_area.selection == Selection.cursor((0, 0))


async def test_insert_text_non_cursor_location_dont_maintain_offset():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.selection = Selection((2, 3), (3, 5))

        result = text_area.insert(
            "Hello",
            location=(4, 0),
            maintain_selection_offset=False,
        )

        assert result == EditResult(
            end_location=(4, 5),
            replaced_text="",
        )
        assert text_area.text == TEXT + "Hello"

        # Since maintain_selection_offset is False, the selection
        # is reset to a cursor and goes to the end of the insert.
        assert text_area.selection == Selection.cursor((4, 5))


async def test_insert_multiline_text():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.move_cursor((2, 5))
        text_area.insert("Hello,\nworld!", maintain_selection_offset=False)
        expected_content = """\
I must not fear.
Fear is the mind-killer.
Fear Hello,
world!is the little-death that brings total obliteration.
I will face my fear.
"""
        assert text_area.cursor_location == (3, 6)  # Cursor moved to end of insert
        assert text_area.text == expected_content


async def test_insert_multiline_text_maintain_offset():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.move_cursor((2, 5))
        result = text_area.insert("Hello,\nworld!")

        assert result == EditResult(
            end_location=(3, 6),
            replaced_text="",
        )

        # The insert happens at the cursor (default location)
        # Offset is maintained - we inserted 1 line so cursor shifts
        # down 1 line, and along by the length of the last insert line.
        assert text_area.cursor_location == (3, 6)
        expected_content = """\
I must not fear.
Fear is the mind-killer.
Fear Hello,
world!is the little-death that brings total obliteration.
I will face my fear.
"""
        assert text_area.text == expected_content


async def test_replace_multiline_text():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        # replace "Fear is the mind-killer\nFear is the little death...\n"
        # with "Hello,\nworld!\n"
        result = text_area.replace("Hello,\nworld!\n", start=(1, 0), end=(3, 0))
        expected_replaced_text = """\
Fear is the mind-killer.
Fear is the little-death that brings total obliteration.
"""
        assert result == EditResult(
            end_location=(3, 0),
            replaced_text=expected_replaced_text,
        )

        expected_content = """\
I must not fear.
Hello,
world!
I will face my fear.
"""
        assert text_area.selection == Selection.cursor((0, 0))  # cursor didnt move
        assert text_area.text == expected_content


async def test_replace_multiline_text_maintain_selection():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)

        # To begin with, the user selects the word "face"
        text_area.selection = Selection((3, 7), (3, 11))
        assert text_area.selected_text == "face"

        # Text is inserted via the API in a way that shifts
        # the start and end locations of the word "face" in
        # both the horizontal and vertical directions.
        text_area.replace(
            "Hello,\nworld!\n123\n456",
            start=(1, 0),
            end=(3, 0),
        )
        expected_content = """\
I must not fear.
Hello,
world!
123
456I will face my fear.
"""
        # Despite this insert, the selection locations are updated
        # and the word face is still highlighted. This ensures that
        # if text is insert programmatically, a user that is typing
        # won't lose their place - the cursor will maintain the same
        # relative position in the document as before.
        assert text_area.selected_text == "face"
        assert text_area.selection == Selection((4, 10), (4, 14))
        assert text_area.text == expected_content


async def test_delete_within_line():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.selection = Selection((0, 11), (0, 15))
        assert text_area.selected_text == "fear"

        # Delete some text before the selection location.
        result = text_area.delete((0, 6), (0, 10))

        # Even though the word has 'shifted' left, it's still selected.
        assert text_area.selection == Selection((0, 7), (0, 11))
        assert text_area.selected_text == "fear"

        # We've recorded exactly what text was replaced in the EditResult
        assert result == EditResult(
            end_location=(0, 6),
            replaced_text=" not",
        )

        expected_text = """\
I must fear.
Fear is the mind-killer.
Fear is the little-death that brings total obliteration.
I will face my fear.
"""
        assert text_area.text == expected_text


async def test_delete_within_line_dont_maintain_offset():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.delete((0, 6), (0, 10), maintain_selection_offset=False)
    expected_text = """\
I must fear.
Fear is the mind-killer.
Fear is the little-death that brings total obliteration.
I will face my fear.
"""
    assert text_area.selection == Selection.cursor((0, 6))  # cursor moved
    assert text_area.text == expected_text


async def test_delete_multiple_lines_selection_above():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)

        # User has selected text on the first line...
        text_area.selection = Selection((0, 2), (0, 6))
        assert text_area.selected_text == "must"

        # Some lines below are deleted...
        result = text_area.delete((1, 0), (3, 0))

        # The selection is not affected at all.
        assert text_area.selection == Selection((0, 2), (0, 6))

        # We've recorded the text that was deleted in the ReplaceResult.
        # Lines of index 1 and 2 were deleted. Since the end
        # location of the selection is (3, 0), the newline
        # marker is included in the deletion.
        expected_replaced_text = """\
Fear is the mind-killer.
Fear is the little-death that brings total obliteration.
"""
        assert result == EditResult(
            end_location=(1, 0),
            replaced_text=expected_replaced_text,
        )
        assert (
            text_area.text
            == """\
I must not fear.
I will face my fear.
"""
        )


async def test_delete_empty_document():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.load_text("")
        result = text_area.delete((0, 0), (1, 0))
        assert result.replaced_text == ""
        assert text_area.text == ""


async def test_clear():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.clear()


async def test_clear_empty_document():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.load_text("")
        text_area.clear()


@pytest.mark.parametrize(
    "select_from,select_to",
    [
        [(0, 3), (2, 1)],
        [(2, 1), (0, 3)],  # Ensuring independence from selection direction.
    ],
)
async def test_insert_text_multiline_selection_top(select_from, select_to):
    """
    An example to attempt to explain what we're testing here...

    X = edit range, * = character in TextArea, S = selection

    *********XX
    XXXXX***SSS
    SSSSSSSSSSS
    SSSS*******

    If an edit happens at XXXX, we need to ensure that the SSS on the
    same line is adjusted appropriately so that it's still highlighting
    the same characters as before.
    """
    app = TextAreaApp()
    async with app.run_test():
        # ABCDE
        # FGHIJ
        # KLMNO
        # PQRST
        # UVWXY
        # Z
        text_area = app.query_one(TextArea)
        text_area.load_text(SIMPLE_TEXT)
        text_area.selection = Selection(select_from, select_to)

        # Check what text is selected.
        expected_selected_text = "DE\nFGHIJ\nK"
        assert text_area.selected_text == expected_selected_text

        result = text_area.replace(
            "Hello",
            start=(0, 0),
            end=(0, 2),
        )

        assert result == EditResult(end_location=(0, 5), replaced_text="AB")

        # The edit range has grown from width 2 to width 5, so the
        # top line of the selection was adjusted (column+=3) such that the
        # same characters are highlighted:
        # ... the selection is not changed after programmatic insert
        # ... the same text is selected as before.
        assert text_area.selected_text == expected_selected_text

        # The resulting text in the TextArea is correct.
        assert text_area.text == "HelloCDE\nFGHIJ\nKLMNO\nPQRST\nUVWXY\nZ\n"


@pytest.mark.parametrize(
    "select_from,select_to",
    [
        [(0, 3), (2, 5)],
        [(2, 5), (0, 3)],  # Ensuring independence from selection direction.
    ],
)
async def test_insert_text_multiline_selection_bottom(select_from, select_to):
    """
    The edited text is within the selected text on the bottom line
    of the selection. The bottom of the selection should be adjusted
    such that any text that was previously selected is still selected.
    """
    app = TextAreaApp()
    async with app.run_test():
        # ABCDE
        # FGHIJ
        # KLMNO
        # PQRST
        # UVWXY
        # Z

        text_area = app.query_one(TextArea)
        text_area.load_text(SIMPLE_TEXT)
        text_area.selection = Selection(select_from, select_to)

        # Check what text is selected.
        assert text_area.selected_text == "DE\nFGHIJ\nKLMNO"

        result = text_area.replace(
            "*",
            start=(2, 0),
            end=(2, 3),
        )
        assert result == EditResult(end_location=(2, 1), replaced_text="KLM")

        # The 'NO' from the selection is still available on the
        # bottom selection line, however the 'KLM' is replaced
        # with '*'. Since 'NO' is still available, it's maintained
        # within the selection.
        assert text_area.selected_text == "DE\nFGHIJ\n*NO"

        # The resulting text in the TextArea is correct.
        # 'KLM' replaced with '*'
        assert text_area.text == "ABCDE\nFGHIJ\n*NO\nPQRST\nUVWXY\nZ\n"


async def test_delete_fully_within_selection():
    """User-facing selection should be best-effort adjusted when a programmatic
    replacement is made to the document."""
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.load_text("0123456789")
        text_area.selection = Selection((0, 2), (0, 7))
        assert text_area.selected_text == "23456"

        result = text_area.delete((0, 4), (0, 6))
        assert result == EditResult(
            replaced_text="45",
            end_location=(0, 4),
        )
        # We deleted 45, but the other characters are still available
        assert text_area.selected_text == "236"
        assert text_area.text == "01236789"


async def test_replace_fully_within_selection():
    """Adjust the selection when a replacement happens inside it."""
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.load_text("0123456789")
        text_area.selection = Selection((0, 2), (0, 7))
        assert text_area.selected_text == "23456"

        result = text_area.replace("XX", start=(0, 2), end=(0, 5))
        assert result == EditResult(
            replaced_text="234",
            end_location=(0, 4),
        )
        assert text_area.selected_text == "XX56"


async def test_text_setter():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        new_text = "hello\nworld\n"
        text_area.text = new_text
        assert text_area.text == new_text


async def test_edits_on_read_only_mode():
    """API edits should still be permitted on read-only mode."""
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.text = "0123456789"
        text_area.read_only = True

        text_area.replace("X", (0, 1), (0, 5))
        assert text_area.text == "0X56789"

        text_area.insert("X")
        assert text_area.text == "X0X56789"

        text_area.delete((0, 0), (0, 2))
        assert text_area.text == "X56789"



from rich.console import RenderableType

from textual.app import App, ComposeResult
from textual.containers import Horizontal
from textual.widget import Widget
from textual.widgets import OptionList


class SimpleLoadingIndicator(Widget):
    """A loading indicator that doesn't animate."""

    DEFAULT_CSS = """
    SimpleLoadingIndicator {
        width: 100%;
        height: 100%;
        min-height: 1;
        content-align: center middle;
        color: $accent;        
    }
    SimpleLoadingIndicator.-textual-loading-indicator {
        layer: _loading;
        background: $boost;     
        dock: top;
    }
    """

    def render(self) -> RenderableType:
        return "Loading!"


class LoadingOverlayRedux(App[None]):
    CSS = """
    OptionList {
        scrollbar-gutter: stable;
        width: 1fr;
        height: 1fr;
    }
    """

    BINDINGS = [("space", "toggle")]

    def get_loading_widget(self) -> Widget:
        return SimpleLoadingIndicator()

    def compose(self) -> ComposeResult:
        with Horizontal():
            yield OptionList(*[("hello world " * 5) for _ in range(100)])
            yield OptionList(*[("foo bar" * 5) for _ in range(100)])

    def action_toggle(self) -> None:
        option_list = self.query(OptionList).first()
        option_list.loading = not option_list.loading


if __name__ == "__main__":
    LoadingOverlayRedux().run()



from textual.app import App, ComposeResult
from textual.widgets import Label, Static


class Box(Static):
    DEFAULT_CSS = """    
    Box {
        border: solid white;
        background: darkblue;
        width: 16;
        height: auto;
    }    
    """

    def compose(self) -> ComposeResult:
        yield Label("FOO\nBAR\nBAZ")


class OffsetsApp(App):
    CSS = """
    
    #box1 {
        offset: 5 5;
    }

    #box2 {
        offset: 15 10;
    }    
    
    """

    def compose(self) -> ComposeResult:
        yield Box(id="box1")
        yield Box(id="box2")


if __name__ == "__main__":
    app = OffsetsApp()
    app.run()



from pathlib import Path

import pytest
from rich.text import Text

import textual.widgets._markdown as MD
from textual.app import App, ComposeResult
from textual.geometry import Offset
from textual.widgets import Markdown, MarkdownViewer, Tree

TEST_MARKDOWN = """\
* [First]({{file}}#first)
* [Second](#second)

# First

The first.

# Second

The second.
"""


class MarkdownFileViewerApp(App[None]):
    def __init__(self, markdown_file: Path) -> None:
        super().__init__()
        self.markdown_file = markdown_file
        markdown_file.write_text(TEST_MARKDOWN.replace("{{file}}", markdown_file.name))

    def compose(self) -> ComposeResult:
        yield MarkdownViewer()

    async def on_mount(self) -> None:
        self.query_one(MarkdownViewer).show_table_of_contents = False
        await self.query_one(MarkdownViewer).go(self.markdown_file)


@pytest.mark.parametrize("link", [0, 1])
async def test_markdown_file_viewer_anchor_link(tmp_path, link: int) -> None:
    """Test https://github.com/Textualize/textual/issues/3094"""
    async with MarkdownFileViewerApp(Path(tmp_path) / "test.md").run_test() as pilot:
        # There's not really anything to test *for* here, but the lack of an
        # exception is the win (before the fix this is testing it would have
        # been FileNotFoundError).
        await pilot.click(Markdown, Offset(2, link))


class MarkdownStringViewerApp(App[None]):
    def __init__(self, markdown_string: str) -> None:
        self.markdown_string = markdown_string
        super().__init__()

    def compose(self) -> ComposeResult:
        yield MarkdownViewer(self.markdown_string)

    async def on_mount(self) -> None:
        self.query_one(MarkdownViewer).show_table_of_contents = False


@pytest.mark.parametrize("link", [0, 1])
async def test_markdown_string_viewer_anchor_link(link: int) -> None:
    """Test https://github.com/Textualize/textual/issues/3094

    Also https://github.com/Textualize/textual/pull/3244#issuecomment-1710278718."""
    async with MarkdownStringViewerApp(
        TEST_MARKDOWN.replace("{{file}}", "")
    ).run_test() as pilot:
        # There's not really anything to test *for* here, but the lack of an
        # exception is the win (before the fix this is testing it would have
        # been FileNotFoundError).
        await pilot.click(Markdown, Offset(2, link))


@pytest.mark.parametrize("text", ["Hey [[/test]]", "[i]Hey there[/i]"])
async def test_headings_that_look_like_they_contain_markup(text: str) -> None:
    """Regression test for https://github.com/Textualize/textual/issues/3689.

    Things that look like markup are escaped in markdown headings in the table of contents.
    """

    document = f"# {text}"
    async with MarkdownStringViewerApp(document).run_test() as pilot:
        assert pilot.app.query_one(MD.MarkdownH1)._text == Text(text)
        toc_tree = pilot.app.query_one(MD.MarkdownTableOfContents).query_one(Tree)
        # The toc label looks like "I {text}" but the I is styled so we drop it.
        toc_label = toc_tree.root.children[0].label
        _, text_label = toc_label.divide([2])
        assert text_label == Text(text)



from textual.app import App
from textual.containers import Grid
from textual.widgets import Label

TEXT = (
    "I must not fear. Fear is the mind-killer. Fear is the little-death that "
    "brings total obliteration. I will face my fear. I will permit it to pass over "
    "me and through me."
)


class TextAlign(App):
    CSS_PATH = "text_align.tcss"

    def compose(self):
        yield Grid(
            Label("[b]Left aligned[/]\n" + TEXT, id="one"),
            Label("[b]Center aligned[/]\n" + TEXT, id="two"),
            Label("[b]Right aligned[/]\n" + TEXT, id="three"),
            Label("[b]Justified[/]\n" + TEXT, id="four"),
        )


if __name__ == "__main__":
    app = TextAlign()
    app.run()



from textual.app import App, ComposeResult
from textual.screen import Screen
from textual.widgets import Header, Footer, Label
from textual.containers import VerticalScroll, Container


class Overlay(Container):
    def compose(self) -> ComposeResult:
        yield Label("This should float over the top")


class Body1(VerticalScroll):
    def compose(self) -> ComposeResult:
        yield Label("My God! It's full of stars! " * 300)


class Body2(VerticalScroll):
    def compose(self) -> ComposeResult:
        yield Label("My God! It's full of stars! " * 300)


class Good(Screen):
    def compose(self) -> ComposeResult:
        yield Header()
        yield Overlay()
        yield Body1()
        yield Footer()


class Bad(Screen):
    def compose(self) -> ComposeResult:
        yield Overlay()
        yield Header()
        yield Body2()
        yield Footer()


class Layers(App[None]):
    CSS = """
    Screen {
        layers: base higher;
    }

    Overlay {
        layer: higher;
        dock: top;
        width: auto;
        height: auto;
        padding: 2;
        border: solid yellow;
        background: red;
        color: yellow;
    }

    Body2 {
        background: green;
    }
    """

    SCREENS = {"good": Good, "bad": Bad}

    BINDINGS = [("t", "toggle", "Toggle Screen")]

    def on_mount(self):
        self.push_screen("good")

    def action_toggle(self):
        self.switch_screen(
            "bad" if self.screen.__class__.__name__ == "Good" else "good"
        )


if __name__ == "__main__":
    Layers().run()



from textual._parser import Parser


def test_read1():
    class TestParser(Parser[str]):
        """A simple parser that reads a byte at a time from a stream."""

        def parse(self, on_token):
            while True:
                data = yield self.read1()
                if not data:
                    break
                on_token(data)

    test_parser = TestParser()
    test_data = "Where there is a Will there is a way!"

    for size in range(1, len(test_data) + 1):
        # Feed the parser in pieces, first 1 character at a time, then 2, etc
        data = []
        for offset in range(0, len(test_data), size):
            for chunk in test_parser.feed(test_data[offset : offset + size]):
                data.append(chunk)
        # Check we have received all the data in characters, no matter the fee dsize
        assert len(data) == len(test_data)
        assert "".join(data) == test_data


def test_read():
    class TestParser(Parser[str]):
        """A parser that reads chunks of a given size from the stream."""

        def __init__(self, size):
            self.size = size
            super().__init__()

        def parse(self, on_token):
            while True:
                data = yield self.read1()
                if not data:
                    break
                on_token(data)

    test_data = "Where there is a Will there is a way!"

    for read_size in range(1, len(test_data) + 1):
        for size in range(1, len(test_data) + 1):
            test_parser = TestParser(read_size)
            data = []
            for offset in range(0, len(test_data), size):
                for chunk in test_parser.feed(test_data[offset : offset + size]):
                    data.append(chunk)
            assert "".join(data) == test_data



from __future__ import annotations

import asyncio

from textual import events
from textual.driver import Driver
from textual.geometry import Size


class HeadlessDriver(Driver):
    """A do-nothing driver for testing."""

    @property
    def is_headless(self) -> bool:
        """Is the driver running in 'headless' mode?"""
        return True

    def _get_terminal_size(self) -> tuple[int, int]:
        if self._size is not None:
            return self._size
        width: int | None = 80
        height: int | None = 25
        import shutil

        try:
            width, height = shutil.get_terminal_size()
        except (AttributeError, ValueError, OSError):
            try:
                width, height = shutil.get_terminal_size()
            except (AttributeError, ValueError, OSError):
                pass
        width = width or 80
        height = height or 25
        return width, height

    def write(self, data: str) -> None:
        """Write data to the output device.

        Args:
            data: Raw data.
        """
        # Nothing to write as this is a headless driver.

    def start_application_mode(self) -> None:
        """Start application mode."""
        loop = asyncio.get_running_loop()

        def send_size_event() -> None:
            """Send first resize event."""
            terminal_size = self._get_terminal_size()
            width, height = terminal_size
            textual_size = Size(width, height)
            event = events.Resize(textual_size, textual_size)
            asyncio.run_coroutine_threadsafe(
                self._app._post_message(event),
                loop=loop,
            )

        send_size_event()

    def disable_input(self) -> None:
        """Disable further input."""

    def stop_application_mode(self) -> None:
        """Stop application mode, restore state."""
        # Nothing to do



from textual.app import App, ComposeResult
from textual.color import Color
from textual.containers import Horizontal
from textual.reactive import reactive
from textual.widgets import Input, Static


class ComputedApp(App):
    CSS_PATH = "computed01.tcss"

    red = reactive(0)
    green = reactive(0)
    blue = reactive(0)
    color = reactive(Color.parse("transparent"))

    def compose(self) -> ComposeResult:
        yield Horizontal(
            Input("0", placeholder="Enter red 0-255", id="red"),
            Input("0", placeholder="Enter green 0-255", id="green"),
            Input("0", placeholder="Enter blue 0-255", id="blue"),
            id="color-inputs",
        )
        yield Static(id="color")

    def compute_color(self) -> Color:  # (1)!
        return Color(self.red, self.green, self.blue).clamped

    def watch_color(self, color: Color) -> None:  # (2)
        self.query_one("#color").styles.background = color

    def on_input_changed(self, event: Input.Changed) -> None:
        try:
            component = int(event.value)
        except ValueError:
            self.bell()
        else:
            if event.input.id == "red":
                self.red = component
            elif event.input.id == "green":
                self.green = component
            else:
                self.blue = component


if __name__ == "__main__":
    app = ComputedApp()
    app.run()



"""Tests inspired by https://github.com/Textualize/textual/issues/4101"""

from __future__ import annotations

from textual.app import App, ComposeResult
from textual.widgets import OptionList
from textual.widgets.option_list import Option


class OptionListApp(App[None]):
    """Test option list application."""

    def compose(self) -> ComposeResult:
        yield OptionList()


async def test_get_after_add() -> None:
    """It should be possible to get an option by ID after adding."""
    async with OptionListApp().run_test() as pilot:
        option_list = pilot.app.query_one(OptionList)
        option_list.add_option(Option("0", id="0"))
        assert option_list.get_option("0").id == "0"



from __future__ import annotations

from textual.app import App


async def test_dynamic_disabled():
    """Check we can dynamically disable bindings."""
    actions = []

    class DynamicApp(App):
        BINDINGS = [
            ("a", "register('a')", "A"),
            ("b", "register('b')", "B"),
            ("c", "register('c')", "B"),
        ]

        def action_register(self, key: str) -> None:
            actions.append(key)

        def check_action(
            self, action: str, parameters: tuple[object, ...]
        ) -> bool | None:
            if action == "register":
                if parameters == ("b",):
                    return False
                if parameters == ("c",):
                    return None
            return True

    app = DynamicApp()
    async with app.run_test() as pilot:
        await pilot.press("a", "b", "c")
        assert actions == ["a"]



from textual.app import App, ComposeResult
from textual.widgets import Static

TEXT = """\
Here is a [@click='app.bell']link[/] which you can click!
"""


class LinksApp(App):
    CSS_PATH = "links.tcss"

    def compose(self) -> ComposeResult:
        yield Static(TEXT)
        yield Static(TEXT, id="custom")


if __name__ == "__main__":
    app = LinksApp()
    app.run()



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Iterable

from textual.css.tokenizer import Expect, Token, Tokenizer

if TYPE_CHECKING:
    from textual.css.types import CSSLocation

PERCENT = r"-?\d+\.?\d*%"
DECIMAL = r"-?\d+\.?\d*"
COMMA = r"\s*,\s*"
OPEN_BRACE = r"\(\s*"
CLOSE_BRACE = r"\s*\)"

HEX_COLOR = r"\#[0-9a-fA-F]{8}|\#[0-9a-fA-F]{6}|\#[0-9a-fA-F]{4}|\#[0-9a-fA-F]{3}"
RGB_COLOR = rf"rgb{OPEN_BRACE}{DECIMAL}{COMMA}{DECIMAL}{COMMA}{DECIMAL}{CLOSE_BRACE}|rgba{OPEN_BRACE}{DECIMAL}{COMMA}{DECIMAL}{COMMA}{DECIMAL}{COMMA}{DECIMAL}{CLOSE_BRACE}"
HSL_COLOR = rf"hsl{OPEN_BRACE}{DECIMAL}{COMMA}{PERCENT}{COMMA}{PERCENT}{CLOSE_BRACE}|hsla{OPEN_BRACE}{DECIMAL}{COMMA}{PERCENT}{COMMA}{PERCENT}{COMMA}{DECIMAL}{CLOSE_BRACE}"

COMMENT_LINE = r"\# .*$"
COMMENT_START = r"\/\*"
SCALAR = rf"{DECIMAL}(?:fr|%|w|h|vw|vh)"
DURATION = r"\d+\.?\d*(?:ms|s)"
NUMBER = r"\-?\d+\.?\d*"
COLOR = rf"{HEX_COLOR}|{RGB_COLOR}|{HSL_COLOR}"
KEY_VALUE = r"[a-zA-Z_-][a-zA-Z0-9_-]*=[0-9a-zA-Z_\-\/]+"
TOKEN = "[a-zA-Z_][a-zA-Z0-9_-]*"
STRING = r"\".*?\""
VARIABLE_REF = r"\$[a-zA-Z0-9_\-]+"

IDENTIFIER = r"[a-zA-Z_\-][a-zA-Z0-9_\-]*"
SELECTOR_TYPE_NAME = r"[A-Z_][a-zA-Z0-9_]*"
"""Selectors representing Widget type names should start with upper case or '_'.

The fact that a selector starts with an upper case letter or '_' is relevant in the
context of nested CSS to help determine whether xxx:yyy is a declaration + value or a
selector + pseudo-class."""
DECLARATION_NAME = r"[a-z][a-zA-Z0-9_\-]*"
"""Declaration of TCSS rules start with lowercase.

The fact that a declaration starts with a lower case letter is relevant in the context
of nested CSS to help determine whether xxx:yyy is a declaration + value or a selector
+ pseudo-class.
"""

# Values permitted in variable and rule declarations.
DECLARATION_VALUES = {
    "scalar": SCALAR,
    "duration": DURATION,
    "number": NUMBER,
    "color": COLOR,
    "key_value": KEY_VALUE,
    "token": TOKEN,
    "string": STRING,
    "variable_ref": VARIABLE_REF,
}

# The tokenizers "expectation" while at the root/highest level of scope
# in the CSS file. At this level we might expect to see selectors, comments,
# variable definitions etc.
expect_root_scope = Expect(
    "selector or end of file",
    whitespace=r"\s+",
    comment_start=COMMENT_START,
    comment_line=COMMENT_LINE,
    selector_start_id=r"\#" + IDENTIFIER,
    selector_start_class=r"\." + IDENTIFIER,
    selector_start_universal=r"\*",
    selector_start=SELECTOR_TYPE_NAME,
    variable_name=rf"{VARIABLE_REF}:",
    declaration_set_end=r"\}",
).expect_eof(True)

expect_root_nested = Expect(
    "selector or end of file",
    whitespace=r"\s+",
    comment_start=COMMENT_START,
    comment_line=COMMENT_LINE,
    declaration_name=DECLARATION_NAME + r"\:",
    selector_start_id=r"\#" + IDENTIFIER,
    selector_start_class=r"\." + IDENTIFIER,
    selector_start_universal=r"\*",
    selector_start=SELECTOR_TYPE_NAME,
    variable_name=rf"{VARIABLE_REF}:",
    declaration_set_end=r"\}",
    nested=r"\&",
)

# After a variable declaration e.g. "$warning-text: TOKENS;"
#              for tokenizing variable value ------^~~~~~~^
expect_variable_name_continue = Expect(
    "variable value",
    variable_value_end=r"\n|;",
    whitespace=r"\s+",
    comment_start=COMMENT_START,
    comment_line=COMMENT_LINE,
    **DECLARATION_VALUES,
).expect_eof(True)

expect_comment_end = Expect(
    "comment end",
    comment_end=re.escape("*/"),
)

# After we come across a selector in CSS e.g. ".my-class", we may
# find other selectors, pseudo-classes... e.g. ".my-class :hover"
expect_selector_continue = Expect(
    "selector or {",
    whitespace=r"\s+",
    comment_start=COMMENT_START,
    comment_line=COMMENT_LINE,
    pseudo_class=r"\:[a-zA-Z_-]+",
    selector_id=r"\#" + IDENTIFIER,
    selector_class=r"\." + IDENTIFIER,
    selector_universal=r"\*",
    selector=SELECTOR_TYPE_NAME,
    combinator_child=">",
    new_selector=r",",
    declaration_set_start=r"\{",
    declaration_set_end=r"\}",
    nested=r"\&",
).expect_eof(True)

# A rule declaration e.g. "text: red;"
#                          ^---^
expect_declaration = Expect(
    "rule or selector",
    nested=r"\&",
    whitespace=r"\s+",
    comment_start=COMMENT_START,
    comment_line=COMMENT_LINE,
    declaration_name=DECLARATION_NAME + r"\:",
    declaration_set_end=r"\}",
    #
    selector_start_id=r"\#" + IDENTIFIER,
    selector_start_class=r"\." + IDENTIFIER,
    selector_start_universal=r"\*",
    selector_start=SELECTOR_TYPE_NAME,
)

expect_declaration_solo = Expect(
    "rule declaration",
    whitespace=r"\s+",
    comment_start=COMMENT_START,
    comment_line=COMMENT_LINE,
    declaration_name=DECLARATION_NAME + r"\:",
    declaration_set_end=r"\}",
).expect_eof(True)

# The value(s)/content from a rule declaration e.g. "text: red;"
#                                                         ^---^
expect_declaration_content = Expect(
    "rule value or end of declaration",
    declaration_end=r";",
    whitespace=r"\s+",
    comment_start=COMMENT_START,
    comment_line=COMMENT_LINE,
    **DECLARATION_VALUES,
    important=r"\!important",
    comma=",",
    declaration_set_end=r"\}",
)

expect_declaration_content_solo = Expect(
    "rule value or end of declaration",
    declaration_end=r";",
    whitespace=r"\s+",
    comment_start=COMMENT_START,
    comment_line=COMMENT_LINE,
    **DECLARATION_VALUES,
    important=r"\!important",
    comma=",",
    declaration_set_end=r"\}",
).expect_eof(True)


class TokenizerState:
    """State machine for the tokenizer.

    Attributes:
        EXPECT: The initial expectation of the tokenizer. Since we start tokenizing
            at the root scope, we might expect to see either a variable or selector, for example.
        STATE_MAP: Maps token names to Expects, defines the sets of valid tokens
            that we'd expect to see next, given the current token. For example, if
            we've just processed a variable declaration name, we next expect to see
            the value of that variable.
    """

    EXPECT = expect_root_scope
    STATE_MAP = {
        "variable_name": expect_variable_name_continue,
        "variable_value_end": expect_root_scope,
        "selector_start": expect_selector_continue,
        "selector_start_id": expect_selector_continue,
        "selector_start_class": expect_selector_continue,
        "selector_start_universal": expect_selector_continue,
        "selector_id": expect_selector_continue,
        "selector_class": expect_selector_continue,
        "selector_universal": expect_selector_continue,
        "declaration_set_start": expect_declaration,
        "declaration_name": expect_declaration_content,
        "declaration_end": expect_declaration,
        "declaration_set_end": expect_root_nested,
        "nested": expect_selector_continue,
    }

    def __call__(self, code: str, read_from: CSSLocation) -> Iterable[Token]:
        tokenizer = Tokenizer(code, read_from=read_from)
        expect = self.EXPECT
        get_token = tokenizer.get_token
        get_state = self.STATE_MAP.get
        nest_level = 0
        while True:
            token = get_token(expect)
            name = token.name
            if name == "comment_line":
                continue
            elif name == "comment_start":
                tokenizer.skip_to(expect_comment_end)
                continue
            elif name == "eof":
                break
            elif name == "declaration_set_start":
                nest_level += 1
            elif name == "declaration_set_end":
                nest_level -= 1
                expect = expect_declaration if nest_level else expect_root_scope
                yield token
                continue
            expect = get_state(name, expect)
            yield token


class DeclarationTokenizerState(TokenizerState):
    EXPECT = expect_declaration_solo
    STATE_MAP = {
        "declaration_name": expect_declaration_content,
        "declaration_end": expect_declaration_solo,
    }


class ValueTokenizerState(TokenizerState):
    EXPECT = expect_declaration_content_solo


tokenize = TokenizerState()
tokenize_declarations = DeclarationTokenizerState()
tokenize_value = ValueTokenizerState()


def tokenize_values(values: dict[str, str]) -> dict[str, list[Token]]:
    """Tokenizes the values in a dict of strings.

    Args:
        values: A mapping of CSS variable name on to a value, to be
            added to the CSS context.

    Returns:
        A mapping of name on to a list of tokens,
    """
    value_tokens = {
        name: list(tokenize_value(value, ("__name__", "")))
        for name, value in values.items()
    }
    return value_tokens



"""
Tests for the tabs underline animation, which is considered a basic animation.
(An animation that also plays on the level BASIC.)
"""

from textual.app import App, ComposeResult
from textual.widgets import Label, TabbedContent, Tabs


class TabbedContentApp(App[None]):
    def compose(self) -> ComposeResult:
        with TabbedContent():
            for _ in range(10):
                yield Label("Hey!")


async def test_tabs_underline_animates_on_full() -> None:
    """The underline takes some time to move when animated."""
    app = TabbedContentApp()
    app.animation_level = "full"

    animations: list[str] = []

    async with app.run_test() as pilot:
        animator = app.animator
        animator._record_animation = animations.append
        app.query_one(Tabs).action_previous_tab()
        await pilot.pause()
        assert "highlight_start" in animations
        assert "highlight_end" in animations


async def test_tabs_underline_animates_on_basic() -> None:
    """The underline takes some time to move when animated."""
    app = TabbedContentApp()
    app.animation_level = "basic"

    animations: list[str] = []

    async with app.run_test() as pilot:
        animator = app.animator
        animator._record_animation = animations.append
        app.query_one(Tabs).action_previous_tab()
        await pilot.pause()
        assert "highlight_start" in animations
        assert "highlight_end" in animations


async def test_tabs_underline_does_not_animate_on_none() -> None:
    """The underline jumps to its final position when not animated."""
    app = TabbedContentApp()
    app.animation_level = "none"

    animations: list[str] = []

    async with app.run_test() as pilot:
        animator = app.animator
        animator._record_animation = animations.append
        app.query_one(Tabs).action_previous_tab()
        await pilot.pause()
        assert "highlight_start" not in animations
        assert "highlight_end" not in animations



"""Test removing options from an option list."""

from __future__ import annotations

import pytest

from textual.app import App, ComposeResult
from textual.geometry import Offset
from textual.widgets import OptionList
from textual.widgets.option_list import Option, OptionDoesNotExist


class OptionListApp(App[None]):
    """Test option list application."""

    def compose(self) -> ComposeResult:
        yield OptionList(
            Option("0", id="0"),
            Option("1", id="1"),
        )


async def test_remove_first_option_via_index() -> None:
    """It should be possible to remove the first option of an option list, via index."""
    async with OptionListApp().run_test() as pilot:
        option_list = pilot.app.query_one(OptionList)
        assert option_list.option_count == 2
        assert option_list.highlighted == 0
        option_list.remove_option_at_index(0)
        assert option_list.option_count == 1
        assert option_list.highlighted == 0


async def test_remove_first_option_via_id() -> None:
    """It should be possible to remove the first option of an option list, via ID."""
    async with OptionListApp().run_test() as pilot:
        option_list = pilot.app.query_one(OptionList)
        assert option_list.option_count == 2
        assert option_list.highlighted == 0
        option_list.remove_option("0")
        assert option_list.option_count == 1
        assert option_list.highlighted == 0


async def test_remove_last_option_via_index() -> None:
    """It should be possible to remove the last option of an option list, via index."""
    async with OptionListApp().run_test() as pilot:
        option_list = pilot.app.query_one(OptionList)
        assert option_list.option_count == 2
        assert option_list.highlighted == 0
        option_list.remove_option_at_index(1)
        assert option_list.option_count == 1
        assert option_list.highlighted == 0


async def test_remove_last_option_via_id() -> None:
    """It should be possible to remove the last option of an option list, via ID."""
    async with OptionListApp().run_test() as pilot:
        option_list = pilot.app.query_one(OptionList)
        assert option_list.option_count == 2
        assert option_list.highlighted == 0
        option_list.remove_option("1")
        assert option_list.option_count == 1
        assert option_list.highlighted == 0


async def test_remove_all_options_via_index() -> None:
    """It should be possible to remove all options via index."""
    async with OptionListApp().run_test() as pilot:
        option_list = pilot.app.query_one(OptionList)
        assert option_list.option_count == 2
        assert option_list.highlighted == 0
        option_list.remove_option_at_index(0)
        option_list.remove_option_at_index(0)
        assert option_list.option_count == 0
        assert option_list.highlighted is None


async def test_remove_all_options_via_id() -> None:
    """It should be possible to remove all options via ID."""
    async with OptionListApp().run_test() as pilot:
        option_list = pilot.app.query_one(OptionList)
        assert option_list.option_count == 2
        assert option_list.highlighted == 0
        option_list.remove_option("0")
        option_list.remove_option("1")
        assert option_list.option_count == 0
        assert option_list.highlighted is None


async def test_remove_invalid_id() -> None:
    """Attempting to remove an option ID that doesn't exist should raise an exception."""
    async with OptionListApp().run_test() as pilot:
        with pytest.raises(OptionDoesNotExist):
            pilot.app.query_one(OptionList).remove_option("does-not-exist")


async def test_remove_invalid_index() -> None:
    """Attempting to remove an option index that doesn't exist should raise an exception."""
    async with OptionListApp().run_test() as pilot:
        with pytest.raises(OptionDoesNotExist):
            pilot.app.query_one(OptionList).remove_option_at_index(23)


async def test_remove_with_hover_on_last_option():
    """https://github.com/Textualize/textual/issues/3270"""
    async with OptionListApp().run_test() as pilot:
        await pilot.hover(OptionList, Offset(1, 1) + Offset(2, 1))
        option_list = pilot.app.query_one(OptionList)
        assert option_list._mouse_hovering_over == 1
        option_list.remove_option_at_index(0)
        assert option_list._mouse_hovering_over == None



from datetime import datetime

from pytz import timezone

from textual.app import App, ComposeResult
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Digits, Label


class WorldClock(Widget):

    time: reactive[datetime] = reactive(datetime.now)

    def __init__(self, timezone: str) -> None:
        self.timezone = timezone
        super().__init__()

    def compose(self) -> ComposeResult:
        yield Label(self.timezone)
        yield Digits()

    def watch_time(self, time: datetime) -> None:
        localized_time = time.astimezone(timezone(self.timezone))
        self.query_one(Digits).update(localized_time.strftime("%H:%M:%S"))


class WorldClockApp(App):
    CSS_PATH = "world_clock01.tcss"

    time: reactive[datetime] = reactive(datetime.now)

    def compose(self) -> ComposeResult:
        yield WorldClock("Europe/London").data_bind(WorldClockApp.time)  # (1)!
        yield WorldClock("Europe/Paris").data_bind(WorldClockApp.time)
        yield WorldClock("Asia/Tokyo").data_bind(WorldClockApp.time)

    def update_time(self) -> None:
        self.time = datetime.now()

    def on_mount(self) -> None:
        self.update_time()
        self.set_interval(1, self.update_time)


if __name__ == "__main__":
    WorldClockApp().run()



from __future__ import annotations

from time import time
from typing import TYPE_CHECKING

from rich.style import Style
from rich.text import Text

if TYPE_CHECKING:
    from textual.app import RenderResult
from textual.color import Gradient
from textual.events import Mount
from textual.widget import Widget


class LoadingIndicator(Widget):
    """Display an animated loading indicator."""

    DEFAULT_CSS = """
    LoadingIndicator {
        width: 100%;
        height: 100%;
        min-height: 1;
        content-align: center middle;
        color: $accent;
    }
    LoadingIndicator.-textual-loading-indicator {
        layer: _loading;
        background: $boost;
        dock: top;
    }
    """

    def __init__(
        self,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ):
        """Initialize a loading indicator.

        Args:
            name: The name of the widget.
            id: The ID of the widget in the DOM.
            classes: The CSS classes for the widget.
            disabled: Whether the widget is disabled or not.
        """
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)

        self._start_time: float = 0.0
        """The time the loading indicator was mounted (a Unix timestamp)."""

    def _on_mount(self, _: Mount) -> None:
        self._start_time = time()
        self.auto_refresh = 1 / 16

    def render(self) -> RenderResult:
        if self.app.animation_level == "none":
            return Text("Loading...")

        elapsed = time() - self._start_time
        speed = 0.8
        dot = "\u25cf"
        _, _, background, color = self.colors

        gradient = Gradient(
            (0.0, background.blend(color, 0.1)),
            (0.7, color),
            (1.0, color.lighten(0.1)),
        )

        blends = [(elapsed * speed - dot_number / 8) % 1 for dot_number in range(5)]

        dots = [
            (
                f"{dot} ",
                Style.from_color(gradient.get_color((1 - blend) ** 2).rich_color),
            )
            for blend in blends
        ]
        indicator = Text.assemble(*dots)
        indicator.rstrip()
        return indicator



from __future__ import annotations

import asyncio
import re
from functools import partial
from pathlib import Path, PurePath
from typing import Callable, Iterable, Optional
from urllib.parse import unquote

from markdown_it import MarkdownIt
from markdown_it.token import Token
from rich import box
from rich.style import Style
from rich.syntax import Syntax
from rich.table import Table
from rich.text import Text
from typing_extensions import TypeAlias

from textual._slug import TrackedSlugs
from textual.app import ComposeResult
from textual.await_complete import AwaitComplete
from textual.containers import Horizontal, Vertical, VerticalScroll
from textual.events import Mount
from textual.message import Message
from textual.reactive import reactive, var
from textual.widget import Widget
from textual.widgets import Static, Tree

TableOfContentsType: TypeAlias = "list[tuple[int, str, str | None]]"
"""Information about the table of contents of a markdown document.

The triples encode the level, the label, and the optional block id of each heading.
"""


class Navigator:
    """Manages a stack of paths like a browser."""

    def __init__(self) -> None:
        self.stack: list[Path] = []
        self.index = 0

    @property
    def location(self) -> Path:
        """The current location.

        Returns:
            A path for the current document.
        """
        if not self.stack:
            return Path(".")
        return self.stack[self.index]

    @property
    def start(self) -> bool:
        """Is the current location at the start of the stack?"""
        return self.index == 0

    @property
    def end(self) -> bool:
        """Is the current location at the end of the stack?"""
        return self.index >= len(self.stack) - 1

    def go(self, path: str | PurePath) -> Path:
        """Go to a new document.

        Args:
            path: Path to new document.

        Returns:
            New location.
        """
        location, anchor = Markdown.sanitize_location(str(path))
        if location == Path(".") and anchor:
            current_file, _ = Markdown.sanitize_location(str(self.location))
            path = f"{current_file}#{anchor}"
        new_path = self.location.parent / Path(path)
        self.stack = self.stack[: self.index + 1]
        new_path = new_path.absolute()
        self.stack.append(new_path)
        self.index = len(self.stack) - 1
        return new_path

    def back(self) -> bool:
        """Go back in the stack.

        Returns:
            True if the location changed, otherwise False.
        """
        if self.index:
            self.index -= 1
            return True
        return False

    def forward(self) -> bool:
        """Go forward in the stack.

        Returns:
            True if the location changed, otherwise False.
        """
        if self.index < len(self.stack) - 1:
            self.index += 1
            return True
        return False


class MarkdownBlock(Static):
    """The base class for a Markdown Element."""

    DEFAULT_CSS = """
    MarkdownBlock {
        height: auto;
    }
    """

    def __init__(self, markdown: Markdown, *args, **kwargs) -> None:
        self._markdown: Markdown = markdown
        """A reference to the Markdown document that contains this block."""
        self._text = Text()
        self._token: Token | None = None
        self._blocks: list[MarkdownBlock] = []
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:
        yield from self._blocks
        self._blocks.clear()

    def set_content(self, text: Text) -> None:
        self._text = text
        self.update(text)

    async def action_link(self, href: str) -> None:
        """Called on link click."""
        self.post_message(Markdown.LinkClicked(self._markdown, href))

    def notify_style_update(self) -> None:
        """If CSS was reloaded, try to rebuild this block from its token."""
        super().notify_style_update()
        self.rebuild()

    def rebuild(self) -> None:
        """Rebuild the content of the block if we have a source token."""
        if self._token is not None:
            self.build_from_token(self._token)

    def build_from_token(self, token: Token) -> None:
        """Build the block content from its source token.

        This method allows the block to be rebuilt on demand, which is useful
        when the styles assigned to the
        [Markdown.COMPONENT_CLASSES][textual.widgets.Markdown.COMPONENT_CLASSES]
        change.

        See https://github.com/Textualize/textual/issues/3464 for more information.

        Args:
            token: The token from which this block is built.
        """

        self._token = token
        style_stack: list[Style] = [Style()]
        content = Text()
        if token.children:
            for child in token.children:
                if child.type == "text":
                    content.append(
                        # Ensure repeating spaces and/or tabs get squashed
                        # down to a single space.
                        re.sub(r"\s+", " ", child.content),
                        style_stack[-1],
                    )
                if child.type == "hardbreak":
                    content.append("\n")
                if child.type == "softbreak":
                    content.append(" ", style_stack[-1])
                elif child.type == "code_inline":
                    content.append(
                        child.content,
                        style_stack[-1]
                        + self._markdown.get_component_rich_style(
                            "code_inline", partial=True
                        ),
                    )
                elif child.type == "em_open":
                    style_stack.append(
                        style_stack[-1]
                        + self._markdown.get_component_rich_style("em", partial=True)
                    )
                elif child.type == "strong_open":
                    style_stack.append(
                        style_stack[-1]
                        + self._markdown.get_component_rich_style(
                            "strong", partial=True
                        )
                    )
                elif child.type == "s_open":
                    style_stack.append(
                        style_stack[-1]
                        + self._markdown.get_component_rich_style("s", partial=True)
                    )
                elif child.type == "link_open":
                    href = child.attrs.get("href", "")
                    action = f"link({href!r})"
                    style_stack.append(
                        style_stack[-1] + Style.from_meta({"@click": action})
                    )
                elif child.type == "image":
                    href = child.attrs.get("src", "")
                    alt = child.attrs.get("alt", "")

                    action = f"link({href!r})"
                    style_stack.append(
                        style_stack[-1] + Style.from_meta({"@click": action})
                    )

                    content.append("🖼  ", style_stack[-1])
                    if alt:
                        content.append(f"({alt})", style_stack[-1])
                    if child.children is not None:
                        for grandchild in child.children:
                            content.append(grandchild.content, style_stack[-1])

                    style_stack.pop()

                elif child.type.endswith("_close"):
                    style_stack.pop()

        self.set_content(content)


class MarkdownHeader(MarkdownBlock):
    """Base class for a Markdown header."""

    DEFAULT_CSS = """
    MarkdownHeader {
        color: $text;
        margin: 2 0 1 0;

    }
    """


class MarkdownH1(MarkdownHeader):
    """An H1 Markdown header."""

    DEFAULT_CSS = """

    MarkdownH1 {
        content-align: center middle;
        text-style: bold;
        color: $success;
        &:light {color: $primary;}
    }
    """


class MarkdownH2(MarkdownHeader):
    """An H2 Markdown header."""

    DEFAULT_CSS = """

    MarkdownH2 {
        text-style: underline;
        color: $success;
        &:light {color: $primary;}
    }
    """


class MarkdownH3(MarkdownHeader):
    """An H3 Markdown header."""

    DEFAULT_CSS = """
    MarkdownH3 {
        text-style: bold;
        color: $success;
        margin: 1 0;
        width: auto;
        &:light {color: $primary;}
    }
    """


class MarkdownH4(MarkdownHeader):
    """An H4 Markdown header."""

    DEFAULT_CSS = """
    MarkdownH4 {
        text-style: bold underline;
        margin: 1 0;
        color: $text;
    }
    """


class MarkdownH5(MarkdownHeader):
    """An H5 Markdown header."""

    DEFAULT_CSS = """
    MarkdownH5 {
        text-style: bold;
        color: $text;
        margin: 1 0;

    }
    """


class MarkdownH6(MarkdownHeader):
    """An H6 Markdown header."""

    DEFAULT_CSS = """
    MarkdownH6 {
        text-style: bold;
        color: $text-muted;
        margin: 1 0;
    }
    """


class MarkdownHorizontalRule(MarkdownBlock):
    """A horizontal rule."""

    DEFAULT_CSS = """
    MarkdownHorizontalRule {
        border-bottom: heavy $primary;
        height: 1;
        padding-top: 1;
        margin-bottom: 1;
    }
    """


class MarkdownParagraph(MarkdownBlock):
    """A paragraph Markdown block."""

    SCOPED_CSS = False
    DEFAULT_CSS = """
    Markdown > MarkdownParagraph {
         margin: 0 0 1 0;
    }
    """


class MarkdownBlockQuote(MarkdownBlock):
    """A block quote Markdown block."""

    DEFAULT_CSS = """
    MarkdownBlockQuote {
        background: $boost;
        border-left: outer $success-darken-2;
        margin: 1 0;
        padding: 0 1;
    }
    MarkdownBlockQuote:light {
        border-left: outer $primary;
    }
    MarkdownBlockQuote > BlockQuote {
        margin-left: 2;
        margin-top: 1;
    }
    """


class MarkdownList(MarkdownBlock):
    DEFAULT_CSS = """

    MarkdownList {
        width: 1fr;
    }

    MarkdownList MarkdownList {
        margin: 0;
        padding-top: 0;
    }
    """


class MarkdownBulletList(MarkdownList):
    """A Bullet list Markdown block."""

    DEFAULT_CSS = """
    MarkdownBulletList {
        margin: 0 0 1 0;
        padding: 0 0;
    }

    MarkdownBulletList Horizontal {
        height: auto;
        width: 1fr;
    }

    MarkdownBulletList Vertical {
        height: auto;
        width: 1fr;
    }
    """

    def compose(self) -> ComposeResult:
        for block in self._blocks:
            if isinstance(block, MarkdownListItem):
                bullet = MarkdownBullet()
                bullet.symbol = block.bullet
                yield Horizontal(bullet, Vertical(*block._blocks))
        self._blocks.clear()


class MarkdownOrderedList(MarkdownList):
    """An ordered list Markdown block."""

    DEFAULT_CSS = """
    MarkdownOrderedList {
        margin: 0 0 1 0;
        padding: 0 0;
    }

    MarkdownOrderedList Horizontal {
        height: auto;
        width: 1fr;
    }

    MarkdownOrderedList Vertical {
        height: auto;
        width: 1fr;
    }
    """

    def compose(self) -> ComposeResult:
        suffix = ". "
        start = 1
        if self._blocks and isinstance(self._blocks[0], MarkdownOrderedListItem):
            try:
                start = int(self._blocks[0].bullet)
            except ValueError:
                pass
        symbol_size = max(
            len(f"{number}{suffix}")
            for number, block in enumerate(self._blocks, start)
            if isinstance(block, MarkdownListItem)
        )
        for number, block in enumerate(self._blocks, start):
            if isinstance(block, MarkdownListItem):
                bullet = MarkdownBullet()
                bullet.symbol = f"{number}{suffix}".rjust(symbol_size + 1)
                yield Horizontal(bullet, Vertical(*block._blocks))

        self._blocks.clear()


class MarkdownTableContent(Widget):
    """Renders a Markdown table."""

    DEFAULT_CSS = """
    MarkdownTableContent {
        width: 100%;
        height: auto;

    }
    MarkdownTableContent > .markdown-table--header {
        text-style: bold;
    }
    """

    COMPONENT_CLASSES = {"markdown-table--header", "markdown-table--lines"}

    def __init__(self, headers: list[Text], rows: list[list[Text]]):
        self.headers = headers
        """List of header text."""
        self.rows = rows
        """The row contents."""
        super().__init__()
        self.shrink = True

    def render(self) -> Table:
        table = Table(
            expand=True,
            box=box.SIMPLE_HEAVY,
            style=self.rich_style,
            header_style=self.get_component_rich_style("markdown-table--header"),
            border_style=self.get_component_rich_style("markdown-table--lines"),
            collapse_padding=True,
            padding=0,
        )
        for header in self.headers:
            table.add_column(header)
        for row in self.rows:
            if row:
                table.add_row(*row)
        return table

    async def action_link(self, href: str) -> None:
        """Pass a link action on to the MarkdownTable parent."""
        if isinstance(self.parent, MarkdownTable):
            await self.parent.action_link(href)


class MarkdownTable(MarkdownBlock):
    """A Table markdown Block."""

    DEFAULT_CSS = """
    MarkdownTable {
        width: 100%;
        background: $panel;
    }
    """

    def compose(self) -> ComposeResult:
        def flatten(block: MarkdownBlock) -> Iterable[MarkdownBlock]:
            for block in block._blocks:
                if block._blocks:
                    yield from flatten(block)
                yield block

        headers: list[Text] = []
        rows: list[list[Text]] = []
        for block in flatten(self):
            if isinstance(block, MarkdownTH):
                headers.append(block._text)
            elif isinstance(block, MarkdownTR):
                rows.append([])
            elif isinstance(block, MarkdownTD):
                rows[-1].append(block._text)

        yield MarkdownTableContent(headers, rows)
        self._blocks.clear()


class MarkdownTBody(MarkdownBlock):
    """A table body Markdown block."""


class MarkdownTHead(MarkdownBlock):
    """A table head Markdown block."""


class MarkdownTR(MarkdownBlock):
    """A table row Markdown block."""


class MarkdownTH(MarkdownBlock):
    """A table header Markdown block."""


class MarkdownTD(MarkdownBlock):
    """A table data Markdown block."""


class MarkdownBullet(Widget):
    """A bullet widget."""

    DEFAULT_CSS = """
    MarkdownBullet {
        width: auto;
        color: $success;
        text-style: bold;
        &:light {
            color: $primary;
        }
    }
    """

    symbol = reactive("\u25cf")
    """The symbol for the bullet."""

    def render(self) -> Text:
        return Text(self.symbol)


class MarkdownListItem(MarkdownBlock):
    """A list item Markdown block."""

    DEFAULT_CSS = """
    MarkdownListItem {
        layout: horizontal;
        margin-right: 1;
        height: auto;
    }

    MarkdownListItem > Vertical {
        width: 1fr;
        height: auto;
    }
    """

    def __init__(self, markdown: Markdown, bullet: str) -> None:
        self.bullet = bullet
        super().__init__(markdown)


class MarkdownOrderedListItem(MarkdownListItem):
    pass


class MarkdownUnorderedListItem(MarkdownListItem):
    pass


class MarkdownFence(MarkdownBlock):
    """A fence Markdown block."""

    DEFAULT_CSS = """
    MarkdownFence {
        margin: 1 0;
        overflow: auto;
        width: 100%;
        height: auto;
        max-height: 20;
        color: rgb(210,210,210);


    }

    MarkdownFence > * {
        width: auto;
    }
    """

    def __init__(self, markdown: Markdown, code: str, lexer: str) -> None:
        super().__init__(markdown)
        self.code = code
        self.lexer = lexer
        self.theme = (
            self._markdown.code_dark_theme
            if self.app.dark
            else self._markdown.code_light_theme
        )

    def _block(self) -> Syntax:
        return Syntax(
            self.code,
            lexer=self.lexer,
            word_wrap=False,
            indent_guides=True,
            padding=(1, 2),
            theme=self.theme,
        )

    def _on_mount(self, _: Mount) -> None:
        """Watch app theme switching."""
        self.watch(self.app, "dark", self._retheme)

    def _retheme(self) -> None:
        """Rerender when the theme changes."""
        self.theme = (
            self._markdown.code_dark_theme
            if self.app.dark
            else self._markdown.code_light_theme
        )
        self.get_child_by_type(Static).update(self._block())

    def compose(self) -> ComposeResult:
        yield Static(
            self._block(),
            expand=True,
            shrink=False,
        )


HEADINGS = {
    "h1": MarkdownH1,
    "h2": MarkdownH2,
    "h3": MarkdownH3,
    "h4": MarkdownH4,
    "h5": MarkdownH5,
    "h6": MarkdownH6,
}

NUMERALS = " ⅠⅡⅢⅣⅤⅥ"


class Markdown(Widget):
    DEFAULT_CSS = """
    Markdown {
        height: auto;
        margin: 0 2 1 2;
        layout: vertical;
        color: $text;
        overflow-y: auto;
    }
    .em {
        text-style: italic;
    }
    .strong {
        text-style: bold;
    }
    .s {
        text-style: strike;
    }
    .code_inline {
        text-style: bold dim;
    }
    """

    COMPONENT_CLASSES = {"em", "strong", "s", "code_inline"}
    """
    These component classes target standard inline markdown styles.
    Changing these will potentially break the standard markdown formatting.

    | Class | Description |
    | :- | :- |
    | `code_inline` | Target text that is styled as inline code. |
    | `em` | Target text that is emphasized inline. |
    | `s` | Target text that is styled inline with strykethrough. |
    | `strong` | Target text that is styled inline with strong. |
    """

    BULLETS = ["\u25cf ", "▪ ", "‣ ", "• ", "⭑ "]

    code_dark_theme: reactive[str] = reactive("material")
    """The theme to use for code blocks when in [dark mode][textual.app.App.dark]."""

    code_light_theme: reactive[str] = reactive("material-light")
    """The theme to use for code blocks when in [light mode][textual.app.App.dark]."""

    def __init__(
        self,
        markdown: str | None = None,
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        parser_factory: Callable[[], MarkdownIt] | None = None,
    ):
        """A Markdown widget.

        Args:
            markdown: String containing Markdown or None to leave blank for now.
            name: The name of the widget.
            id: The ID of the widget in the DOM.
            classes: The CSS classes of the widget.
            parser_factory: A factory function to return a configured MarkdownIt instance. If `None`, a "gfm-like" parser is used.
        """
        super().__init__(name=name, id=id, classes=classes)
        self._markdown = markdown
        self._parser_factory = parser_factory
        self._table_of_contents: TableOfContentsType | None = None

    class TableOfContentsUpdated(Message):
        """The table of contents was updated."""

        def __init__(
            self, markdown: Markdown, table_of_contents: TableOfContentsType
        ) -> None:
            super().__init__()
            self.markdown: Markdown = markdown
            """The `Markdown` widget associated with the table of contents."""
            self.table_of_contents: TableOfContentsType = table_of_contents
            """Table of contents."""

        @property
        def control(self) -> Markdown:
            """The `Markdown` widget associated with the table of contents.

            This is an alias for [`TableOfContentsUpdated.markdown`][textual.widgets.Markdown.TableOfContentsSelected.markdown]
            and is used by the [`on`][textual.on] decorator.
            """
            return self.markdown

    class TableOfContentsSelected(Message):
        """An item in the TOC was selected."""

        def __init__(self, markdown: Markdown, block_id: str) -> None:
            super().__init__()
            self.markdown: Markdown = markdown
            """The `Markdown` widget where the selected item is."""
            self.block_id: str = block_id
            """ID of the block that was selected."""

        @property
        def control(self) -> Markdown:
            """The `Markdown` widget where the selected item is.

            This is an alias for [`TableOfContentsSelected.markdown`][textual.widgets.Markdown.TableOfContentsSelected.markdown]
            and is used by the [`on`][textual.on] decorator.
            """
            return self.markdown

    class LinkClicked(Message):
        """A link in the document was clicked."""

        def __init__(self, markdown: Markdown, href: str) -> None:
            super().__init__()
            self.markdown: Markdown = markdown
            """The `Markdown` widget containing the link clicked."""
            self.href: str = unquote(href)
            """The link that was selected."""

        @property
        def control(self) -> Markdown:
            """The `Markdown` widget containing the link clicked.

            This is an alias for [`LinkClicked.markdown`][textual.widgets.Markdown.LinkClicked.markdown]
            and is used by the [`on`][textual.on] decorator.
            """
            return self.markdown

    async def _on_mount(self, _: Mount) -> None:
        if self._markdown is not None:
            await self.update(self._markdown)

    def _watch_code_dark_theme(self) -> None:
        """React to the dark theme being changed."""
        if self.app.dark:
            for block in self.query(MarkdownFence):
                block._retheme()

    def _watch_code_light_theme(self) -> None:
        """React to the light theme being changed."""
        if not self.app.dark:
            for block in self.query(MarkdownFence):
                block._retheme()

    @staticmethod
    def sanitize_location(location: str) -> tuple[Path, str]:
        """Given a location, break out the path and any anchor.

        Args:
            location: The location to sanitize.

        Returns:
            A tuple of the path to the location cleaned of any anchor, plus
            the anchor (or an empty string if none was found).
        """
        location, _, anchor = location.partition("#")
        return Path(location), anchor

    def goto_anchor(self, anchor: str) -> bool:
        """Try and find the given anchor in the current document.

        Args:
            anchor: The anchor to try and find.

        Note:
            The anchor is found by looking at all of the headings in the
            document and finding the first one whose slug matches the
            anchor.

            Note that the slugging method used is similar to that found on
            GitHub.

        Returns:
            True when the anchor was found in the current document, False otherwise.
        """
        if not self._table_of_contents or not isinstance(self.parent, Widget):
            return False
        unique = TrackedSlugs()
        for _, title, header_id in self._table_of_contents:
            if unique.slug(title) == anchor:
                self.query_one(f"#{header_id}").scroll_visible(top=True)
                return True
        return False

    async def load(self, path: Path) -> None:
        """Load a new Markdown document.

        Args:
            path: Path to the document.

        Raises:
            OSError: If there was some form of error loading the document.

        Note:
            The exceptions that can be raised by this method are all of
            those that can be raised by calling [`Path.read_text`][pathlib.Path.read_text].
        """
        path, anchor = self.sanitize_location(str(path))
        data = await asyncio.get_running_loop().run_in_executor(
            None, partial(path.read_text, encoding="utf-8")
        )
        await self.update(data)
        if anchor:
            self.goto_anchor(anchor)

    def unhandled_token(self, token: Token) -> MarkdownBlock | None:
        """Process an unhandled token.

        Args:
            token: The MarkdownIt token to handle.

        Returns:
            Either a widget to be added to the output, or `None`.
        """
        return None

    def update(self, markdown: str) -> AwaitComplete:
        """Update the document with new Markdown.

        Args:
            markdown: A string containing Markdown.

        Returns:
            An optionally awaitable object. Await this to ensure that all children have been mounted.
        """
        parser = (
            MarkdownIt("gfm-like")
            if self._parser_factory is None
            else self._parser_factory()
        )

        table_of_contents = []

        def parse_markdown(tokens) -> Iterable[MarkdownBlock]:
            """Create a stream of MarkdownBlock widgets from markdown.

            Args:
                tokens: List of tokens

            Yields:
                Widgets for mounting.
            """

            stack: list[MarkdownBlock] = []
            stack_append = stack.append
            block_id: int = 0

            for token in tokens:
                token_type = token.type
                if token_type == "heading_open":
                    block_id += 1
                    stack_append(HEADINGS[token.tag](self, id=f"block{block_id}"))
                elif token_type == "hr":
                    yield MarkdownHorizontalRule(self)
                elif token_type == "paragraph_open":
                    stack_append(MarkdownParagraph(self))
                elif token_type == "blockquote_open":
                    stack_append(MarkdownBlockQuote(self))
                elif token_type == "bullet_list_open":
                    stack_append(MarkdownBulletList(self))
                elif token_type == "ordered_list_open":
                    stack_append(MarkdownOrderedList(self))
                elif token_type == "list_item_open":
                    if token.info:
                        stack_append(MarkdownOrderedListItem(self, token.info))
                    else:
                        item_count = sum(
                            1
                            for block in stack
                            if isinstance(block, MarkdownUnorderedListItem)
                        )
                        stack_append(
                            MarkdownUnorderedListItem(
                                self,
                                self.BULLETS[item_count % len(self.BULLETS)],
                            )
                        )
                elif token_type == "table_open":
                    stack_append(MarkdownTable(self))
                elif token_type == "tbody_open":
                    stack_append(MarkdownTBody(self))
                elif token_type == "thead_open":
                    stack_append(MarkdownTHead(self))
                elif token_type == "tr_open":
                    stack_append(MarkdownTR(self))
                elif token_type == "th_open":
                    stack_append(MarkdownTH(self))
                elif token_type == "td_open":
                    stack_append(MarkdownTD(self))
                elif token_type.endswith("_close"):
                    block = stack.pop()
                    if token.type == "heading_close":
                        heading = block._text.plain
                        level = int(token.tag[1:])
                        table_of_contents.append((level, heading, block.id))
                    if stack:
                        stack[-1]._blocks.append(block)
                    else:
                        yield block
                elif token_type == "inline":
                    stack[-1].build_from_token(token)
                elif token_type in ("fence", "code_block"):
                    fence = MarkdownFence(self, token.content.rstrip(), token.info)
                    if stack:
                        stack[-1]._blocks.append(fence)
                    else:
                        yield fence
                else:
                    external = self.unhandled_token(token)
                    if external is not None:
                        if stack:
                            stack[-1]._blocks.append(external)
                        else:
                            yield external

        markdown_block = self.query("MarkdownBlock")

        async def await_update() -> None:
            """Update in batches."""
            BATCH_SIZE = 200
            batch: list[MarkdownBlock] = []
            tokens = await asyncio.get_running_loop().run_in_executor(
                None, parser.parse, markdown
            )

            # Lock so that you can't update with more than one document simultaneously
            async with self.lock:
                # Remove existing blocks for the first batch only
                removed: bool = False

                async def mount_batch(batch: list[MarkdownBlock]) -> None:
                    """Mount a single match of blocks.

                    Args:
                        batch: A list of blocks to mount.
                    """
                    nonlocal removed
                    if removed:
                        await self.mount_all(batch)
                    else:
                        with self.app.batch_update():
                            await markdown_block.remove()
                            await self.mount_all(batch)
                        removed = True

                for block in parse_markdown(tokens):
                    batch.append(block)
                    if len(batch) == BATCH_SIZE:
                        await mount_batch(batch)
                        batch.clear()
                if batch:
                    await mount_batch(batch)
                if not removed:
                    await markdown_block.remove()

            self._table_of_contents = table_of_contents

            self.post_message(
                Markdown.TableOfContentsUpdated(
                    self, self._table_of_contents
                ).set_sender(self)
            )

        return AwaitComplete(await_update())


class MarkdownTableOfContents(Widget, can_focus_children=True):
    """Displays a table of contents for a markdown document."""

    DEFAULT_CSS = """
    MarkdownTableOfContents {
        width: auto;
        background: $panel;
        border-right: wide $background;
    }
    MarkdownTableOfContents > Tree {
        padding: 1;
        width: auto;
    }
    """

    table_of_contents = reactive[Optional[TableOfContentsType]](None, init=False)
    """Underlying data to populate the table of contents widget."""

    def __init__(
        self,
        markdown: Markdown,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        """Initialize a table of contents.

        Args:
            markdown: The Markdown document associated with this table of contents.
            name: The name of the widget.
            id: The ID of the widget in the DOM.
            classes: The CSS classes for the widget.
            disabled: Whether the widget is disabled or not.
        """
        self.markdown: Markdown = markdown
        """The Markdown document associated with this table of contents."""
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)

    def compose(self) -> ComposeResult:
        tree: Tree = Tree("TOC")
        tree.show_root = False
        tree.show_guides = True
        tree.guide_depth = 4
        tree.auto_expand = False
        yield tree

    def watch_table_of_contents(self, table_of_contents: TableOfContentsType) -> None:
        """Triggered when the table of contents changes."""
        self.rebuild_table_of_contents(table_of_contents)

    def rebuild_table_of_contents(self, table_of_contents: TableOfContentsType) -> None:
        """Rebuilds the tree representation of the table of contents data.

        Args:
            table_of_contents: Table of contents.
        """
        tree = self.query_one(Tree)
        tree.clear()
        root = tree.root
        for level, name, block_id in table_of_contents:
            node = root
            for _ in range(level - 1):
                if node._children:
                    node = node._children[-1]
                    node.expand()
                    node.allow_expand = True
                else:
                    node = node.add(NUMERALS[level], expand=True)
            node_label = Text.assemble((f"{NUMERALS[level]} ", "dim"), name)
            node.add_leaf(node_label, {"block_id": block_id})

    async def _on_tree_node_selected(self, message: Tree.NodeSelected) -> None:
        node_data = message.node.data
        if node_data is not None:
            await self._post_message(
                Markdown.TableOfContentsSelected(self.markdown, node_data["block_id"])
            )
        message.stop()


class MarkdownViewer(VerticalScroll, can_focus=True, can_focus_children=True):
    """A Markdown viewer widget."""

    SCOPED_CSS = False

    DEFAULT_CSS = """
    MarkdownViewer {
        height: 1fr;
        scrollbar-gutter: stable;
    }

    MarkdownTableOfContents {
        dock:left;
    }

    MarkdownViewer > MarkdownTableOfContents {
        display: none;
    }

    MarkdownViewer.-show-table-of-contents > MarkdownTableOfContents {
        display: block;
    }
    """

    show_table_of_contents = reactive(True)
    top_block = reactive("")

    navigator: var[Navigator] = var(Navigator)

    class NavigatorUpdated(Message):
        """Navigator has been changed (clicked link etc)."""

    def __init__(
        self,
        markdown: str | None = None,
        *,
        show_table_of_contents: bool = True,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        parser_factory: Callable[[], MarkdownIt] | None = None,
    ):
        """Create a Markdown Viewer object.

        Args:
            markdown: String containing Markdown, or None to leave blank.
            show_table_of_contents: Show a table of contents in a sidebar.
            name: The name of the widget.
            id: The ID of the widget in the DOM.
            classes: The CSS classes of the widget.
            parser_factory: A factory function to return a configured MarkdownIt instance. If `None`, a "gfm-like" parser is used.
        """
        super().__init__(name=name, id=id, classes=classes)
        self.show_table_of_contents = show_table_of_contents
        self._markdown = markdown
        self._parser_factory = parser_factory

    @property
    def document(self) -> Markdown:
        """The [`Markdown`][textual.widgets.Markdown] document widget."""
        return self.query_one(Markdown)

    @property
    def table_of_contents(self) -> MarkdownTableOfContents:
        """The [table of contents][textual.widgets.markdown.MarkdownTableOfContents] widget."""
        return self.query_one(MarkdownTableOfContents)

    async def _on_mount(self, _: Mount) -> None:
        if self._markdown is not None:
            await self.document.update(self._markdown)

    async def go(self, location: str | PurePath) -> None:
        """Navigate to a new document path."""
        path, anchor = self.document.sanitize_location(str(location))
        if path == Path(".") and anchor:
            # We've been asked to go to an anchor but with no file specified.
            self.document.goto_anchor(anchor)
        else:
            # We've been asked to go to a file, optionally with an anchor.
            await self.document.load(self.navigator.go(location))
            self.post_message(self.NavigatorUpdated())

    async def back(self) -> None:
        """Go back one level in the history."""
        if self.navigator.back():
            await self.document.load(self.navigator.location)
            self.post_message(self.NavigatorUpdated())

    async def forward(self) -> None:
        """Go forward one level in the history."""
        if self.navigator.forward():
            await self.document.load(self.navigator.location)
            self.post_message(self.NavigatorUpdated())

    async def _on_markdown_link_clicked(self, message: Markdown.LinkClicked) -> None:
        message.stop()
        await self.go(message.href)

    def watch_show_table_of_contents(self, show_table_of_contents: bool) -> None:
        self.set_class(show_table_of_contents, "-show-table-of-contents")

    def compose(self) -> ComposeResult:
        markdown = Markdown(parser_factory=self._parser_factory)
        yield MarkdownTableOfContents(markdown)
        yield markdown

    def _on_markdown_table_of_contents_updated(
        self, message: Markdown.TableOfContentsUpdated
    ) -> None:
        self.query_one(MarkdownTableOfContents).table_of_contents = (
            message.table_of_contents
        )
        message.stop()

    def _on_markdown_table_of_contents_selected(
        self, message: Markdown.TableOfContentsSelected
    ) -> None:
        block_selector = f"#{message.block_id}"
        block = self.query_one(block_selector, MarkdownBlock)
        self.scroll_to_widget(block, top=True)
        message.stop()



from textual.app import App, ComposeResult
from textual.widgets import RichLog
from textual.containers import Horizontal


class RichLogScrollApp(App):
    CSS = """
    RichLog{
        width: 1fr;
        height: 10;
    }
    """

    def compose(self) -> ComposeResult:
        with Horizontal():
            # Don't scroll on write
            yield RichLog(id="richlog1", auto_scroll=False)
            # Scroll on write
            yield RichLog(id="richlog2", auto_scroll=True)
            # Scroll on write, but disabled on write()
            yield RichLog(id="richlog3", auto_scroll=True)

    def on_ready(self) -> None:
        lines = [f"Line {n}" for n in range(20)]
        for line in lines:
            self.query_one("#richlog1", RichLog).write(line)
        for line in lines:
            self.query_one("#richlog2", RichLog).write(line)
        for line in lines:
            self.query_one("#richlog3", RichLog).write(line, scroll_end=False)


if __name__ == "__main__":
    app = RichLogScrollApp()
    app.run()



from threading import Thread

import pytest

from textual.app import App, ComposeResult
from textual.widgets import RichLog


def test_call_from_thread_app_not_running():
    app = App()

    # Should fail if app is not running
    with pytest.raises(RuntimeError):
        app.call_from_thread(print)


def test_call_from_thread():
    """Test the call_from_thread method."""

    class BackgroundThread(Thread):
        """A background thread which will modify app in some way."""

        def __init__(self, app: App[object]) -> None:
            self.app = app
            super().__init__()

        def run(self) -> None:
            def write_stuff(text: str) -> None:
                """Write stuff to a widget."""
                self.app.query_one(RichLog).write(text)

            self.app.call_from_thread(write_stuff, "Hello")
            # Exit the app with a code we can assert
            self.app.call_from_thread(self.app.exit, 123)

    class ThreadTestApp(App[object]):
        """Trivial app with a single widget."""

        def compose(self) -> ComposeResult:
            yield RichLog()

        def on_ready(self) -> None:
            """Launch a thread which will modify the app."""
            try:
                self.call_from_thread(print)
            except RuntimeError as error:
                # Calling this from the same thread as the app is an error
                self._runtime_error = error
            BackgroundThread(self).start()

    app = ThreadTestApp()
    result = app.run(headless=True, size=(80, 24))
    assert isinstance(app._runtime_error, RuntimeError)
    assert result == 123



"""Provides an immutable sequence view class."""

from __future__ import annotations

from sys import maxsize
from typing import TYPE_CHECKING, Generic, Iterator, Sequence, TypeVar, overload

T = TypeVar("T")


class ImmutableSequenceView(Generic[T]):
    """Class to wrap a sequence of some sort, but not allow modification."""

    def __init__(self, wrap: Sequence[T]) -> None:
        """Initialise the immutable sequence.

        Args:
            wrap: The sequence being wrapped.
        """
        self._wrap = wrap

    if TYPE_CHECKING:

        @overload
        def __getitem__(self, index: int) -> T: ...

        @overload
        def __getitem__(self, index: slice) -> ImmutableSequenceView[T]: ...

    def __getitem__(self, index: int | slice) -> T | ImmutableSequenceView[T]:
        return (
            self._wrap[index]
            if isinstance(index, int)
            else ImmutableSequenceView[T](self._wrap[index])
        )

    def __iter__(self) -> Iterator[T]:
        return iter(self._wrap)

    def __len__(self) -> int:
        return len(self._wrap)

    def __length_hint__(self) -> int:
        return len(self)

    def __bool__(self) -> bool:
        return bool(self._wrap)

    def __contains__(self, item: T) -> bool:
        return item in self._wrap

    def index(self, item: T, start: int = 0, stop: int = maxsize) -> int:
        """Return the index of the given item.

        Args:
            item: The item to find in the sequence.
            start: Optional start location.
            stop: Optional stop location.

        Returns:
            The index of the item in the sequence.

        Raises:
            ValueError: If the item is not in the sequence.
        """
        return self._wrap.index(item, start, stop)

    def __reversed__(self) -> Iterator[T]:
        return reversed(self._wrap)



from textual.app import App
from textual.containers import Container
from textual.widgets import Label


class DockAllApp(App):
    CSS_PATH = "dock_all.tcss"

    def compose(self):
        yield Container(
            Container(Label("left"), id="left"),
            Container(Label("top"), id="top"),
            Container(Label("right"), id="right"),
            Container(Label("bottom"), id="bottom"),
            id="big_container",
        )


if __name__ == "__main__":
    app = DockAllApp()
    app.run()



from textual.app import App
from textual.widgets import Header, Label, Footer


class TestApp(App):
    BINDINGS = [("ctrl+q", "app.quit", "Quit")]
    CSS = """
    
    Label {
        border: solid red;
    }
    Footer {
        height: 4;
    }
    """

    def compose(self):
        text = (
            "this is a sample sentence and here are some words".replace(" ", "\n") * 2
        )
        yield Header()
        yield Label(text)
        yield Footer()

    def on_mount(self):
        self.dark = False


if __name__ == "__main__":
    app = TestApp()
    app.run()



import asyncio
from time import sleep
from typing import Callable, List, Tuple

import pytest

from textual import work
from textual._work_decorator import WorkerDeclarationError
from textual.app import App
from textual.worker import Worker, WorkerState, WorkType


class WorkApp(App):
    worker: Worker

    def __init__(self) -> None:
        super().__init__()
        self.states: list[WorkerState] = []

    @work
    async def async_work(self) -> str:
        await asyncio.sleep(0.1)
        return "foo"

    @work(thread=True)
    async def async_thread_work(self) -> str:
        await asyncio.sleep(0.1)
        return "foo"

    @work(thread=True)
    def thread_work(self) -> str:
        sleep(0.1)
        return "foo"

    def launch(self, worker) -> None:
        self.worker = worker()

    def on_worker_state_changed(self, event: Worker.StateChanged) -> None:
        self.states.append(event.state)


async def work_with(launcher: Callable[[WorkApp], WorkType]) -> None:
    """Core code for testing a work decorator."""
    app = WorkApp()
    async with app.run_test() as pilot:
        app.launch(launcher(app))
        await app.workers.wait_for_complete()
        result = await app.worker.wait()
        assert result == "foo"
        await pilot.pause()
        assert app.states == [
            WorkerState.PENDING,
            WorkerState.RUNNING,
            WorkerState.SUCCESS,
        ]


async def test_async_work() -> None:
    """It should be possible to decorate an async method as an async worker."""
    await work_with(lambda app: app.async_work)


async def test_async_thread_work() -> None:
    """It should be possible to decorate an async method as a thread worker."""
    await work_with(lambda app: app.async_thread_work)


async def test_thread_work() -> None:
    """It should be possible to decorate a non-async method as a thread worker."""
    await work_with(lambda app: app.thread_work)


def test_decorate_non_async_no_thread_argument() -> None:
    """Decorating a non-async method without saying explicitly that it's a thread is an error."""
    with pytest.raises(WorkerDeclarationError):

        class _(App[None]):
            @work
            def foo(self) -> None:
                pass


def test_decorate_non_async_no_thread_is_false() -> None:
    """Decorating a non-async method and saying it isn't a thread is an error."""
    with pytest.raises(WorkerDeclarationError):

        class _(App[None]):
            @work(thread=False)
            def foo(self) -> None:
                pass


class NestedWorkersApp(App[None]):
    def __init__(self, call_stack: List[str]):
        self.call_stack = call_stack
        super().__init__()

    def call_from_stack(self):
        if self.call_stack:
            call_now = self.call_stack.pop()
            getattr(self, call_now)()

    @work(thread=False)
    async def async_no_thread(self):
        self.call_from_stack()

    @work(thread=True)
    async def async_thread(self):
        self.call_from_stack()

    @work(thread=True)
    def thread(self):
        self.call_from_stack()


@pytest.mark.parametrize(
    "call_stack",
    [  # from itertools import product; list(product("async_no_thread async_thread thread".split(), repeat=3))
        ("async_no_thread", "async_no_thread", "async_no_thread"),
        ("async_no_thread", "async_no_thread", "async_thread"),
        ("async_no_thread", "async_no_thread", "thread"),
        ("async_no_thread", "async_thread", "async_no_thread"),
        ("async_no_thread", "async_thread", "async_thread"),
        ("async_no_thread", "async_thread", "thread"),
        ("async_no_thread", "thread", "async_no_thread"),
        ("async_no_thread", "thread", "async_thread"),
        ("async_no_thread", "thread", "thread"),
        ("async_thread", "async_no_thread", "async_no_thread"),
        ("async_thread", "async_no_thread", "async_thread"),
        ("async_thread", "async_no_thread", "thread"),
        ("async_thread", "async_thread", "async_no_thread"),
        ("async_thread", "async_thread", "async_thread"),
        ("async_thread", "async_thread", "thread"),
        ("async_thread", "thread", "async_no_thread"),
        ("async_thread", "thread", "async_thread"),
        ("async_thread", "thread", "thread"),
        ("thread", "async_no_thread", "async_no_thread"),
        ("thread", "async_no_thread", "async_thread"),
        ("thread", "async_no_thread", "thread"),
        ("thread", "async_thread", "async_no_thread"),
        ("thread", "async_thread", "async_thread"),
        ("thread", "async_thread", "thread"),
        ("thread", "thread", "async_no_thread"),
        ("thread", "thread", "async_thread"),
        ("thread", "thread", "thread"),
        (  # Plus a longer chain to stress test this mechanism.
            "async_no_thread",
            "async_no_thread",
            "thread",
            "thread",
            "async_thread",
            "async_thread",
            "async_no_thread",
            "async_thread",
            "async_no_thread",
            "async_thread",
            "thread",
            "async_thread",
            "async_thread",
            "async_no_thread",
            "async_no_thread",
            "thread",
            "thread",
            "async_no_thread",
            "async_no_thread",
            "thread",
            "async_no_thread",
            "thread",
            "thread",
        ),
    ],
)
async def test_calling_workers_from_within_workers(call_stack: Tuple[str]):
    """Regression test for https://github.com/Textualize/textual/issues/3472.

    This makes sure we can nest worker calls without a problem.
    """
    app = NestedWorkersApp(list(call_stack))
    async with app.run_test():
        app.call_from_stack()
        # We need multiple awaits because we're creating a chain of workers that may
        # have multiple async workers, each of which may need the await to have enough
        # time to call the next one in the chain.
        for _ in range(len(call_stack)):
            await app.workers.wait_for_complete()
        assert app.call_stack == []
